"""Logger."""

import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

file_handler = logging.FileHandler("monitor_log.log")
file_handler.setLevel(logging.DEBUG)

formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
file_handler.setFormatter(formatter)


logger.addHandler(file_handler)



"""Trusted value for EVMCall query type."""

import math
from typing import Any

from layer_values_monitor.logger import logger

from hexbytes import HexBytes
from telliot_core.apps.telliot_config import TelliotConfig
from telliot_feeds.feeds import DataFeed
from web3 import Web3
from web3.exceptions import ExtraDataLengthError
from web3.middleware import geth_poa_middleware


async def get_evm_call_trusted_value(reported_val: Any, feed: DataFeed) -> HexBytes:
    """Get trusted value for EVMCall query type."""
    if not isinstance(reported_val, tuple):
        return True
    block_timestamp = reported_val[1]
    reported_val = HexBytes(reported_val[0])

    block_number = get_block_number_at_timestamp(feed.query.chainId, block_timestamp)

    trusted_val, _ = await feed.source.fetch_new_datapoint(block_number)
    if not isinstance(trusted_val, tuple):
        logger.warning(f"Bad value response for EVMCall: {trusted_val}")
        return None

    if trusted_val[0] is None:
        logger.warning(f"Unable to fetch trusted value for EVMCall: {trusted_val}")
        return None
    return HexBytes(trusted_val[0])


def get_block_number_at_timestamp(chain_id: int, timestamp: int) -> int | None:
    """Get block number for a given timestamp from the identified rpc."""
    cfg = TelliotConfig()
    cfg.main.chain_id = chain_id
    try:
        endpoint = cfg.get_endpoint()
        endpoint.connect()
    except ValueError as e:
        logger.error(f"Unable to connect to endpoint on chain_id {cfg.main.chain_id}: {e}")
        return None

    w3: Web3 = endpoint.web3

    current_block = w3.eth.block_number
    start_block = 0
    end_block = current_block

    while start_block <= end_block:
        midpoint = math.floor((start_block + end_block) / 2)
        # for poa chains get_block method throws an error if poa middleware is not injected
        try:
            block = w3.eth.get_block(midpoint)
        except ExtraDataLengthError:
            w3.middleware_onion.inject(geth_poa_middleware, layer=0)
            block = w3.eth.get_block(midpoint)

        if block.timestamp == timestamp:
            return midpoint
        elif block.timestamp < timestamp:
            start_block = midpoint + 1
        else:
            end_block = midpoint - 1

    # If we haven't found an exact match, interpolate between adjacent blocks
    block_a = w3.eth.get_block(end_block)
    block_b = w3.eth.get_block(start_block)

    block_delta = block_b.number - block_a.number
    timestamp_delta = block_b.timestamp - block_a.timestamp
    target_delta = timestamp - block_a.timestamp

    estimated_block_delta = target_delta * block_delta / timestamp_delta
    estimated_block_number = block_a.number + estimated_block_delta

    return int(estimated_block_number)



"""Send messages using Discord."""

import os

import click
from discordwebhook import Discord


def generic_alert(msg: str) -> None:
    """Send a Discord message via webhook."""
    send_discord_msg(msg)
    return


def get_alert_bot_1() -> Discord:
    """Read the Discord webhook url from the environment."""
    DISCORD_WEBHOOK_URL_1 = os.getenv("DISCORD_WEBHOOK_URL_1")
    if DISCORD_WEBHOOK_URL_1 is None:
        raise Exception("At least one DISCORD_WEBHOOK_URL is required.")
    alert_bot_1 = Discord(url=DISCORD_WEBHOOK_URL_1)
    return alert_bot_1


def get_alert_bot_2() -> Discord:
    """Read the Discord webhook url from the environment."""
    return Discord(url=os.getenv("DISCORD_WEBHOOK_URL_2"))


def get_alert_bot_3() -> Discord:
    """Read the Discord webhook url from the environment."""
    return Discord(url=os.getenv("DISCORD_WEBHOOK_URL_3"))


def send_discord_msg(msg: str) -> None:
    """Send Discord alert."""
    MONITOR_NAME = os.getenv("MONITOR_NAME")
    message = f"❗{MONITOR_NAME} Found Something❗\n"
    get_alert_bot_1().post(content=message + msg)
    try:
        get_alert_bot_2().post(content=message + msg)
    except Exception as e:
        click.echo(f"alert bot 2 not used? {e}")
        pass
    try:
        get_alert_bot_3().post(content=message + msg)
    except Exception as e:
        click.echo(f"alert bot 3 not used? {e}")
        pass
    click.echo("Alerts sent!")
    return



"""Telliot Feeds helper functions."""

import logging

from clamfig.base import Registry
from eth_abi import decode
from telliot_feeds.datafeed import DataFeed
from telliot_feeds.datasource import DataSource
from telliot_feeds.dtypes.datapoint import OptionalDataPoint
from telliot_feeds.feeds import CATALOG_FEEDS, DATAFEED_BUILDER_MAPPING
from telliot_feeds.queries.abi_query import AbiQuery
from telliot_feeds.queries.json_query import JsonQuery
from telliot_feeds.queries.query_catalog import query_catalog


def get_query_from_data(query_data: bytes) -> AbiQuery | JsonQuery | None:
    """Get query give query data from telliot-feeds."""
    for q_type in (JsonQuery, AbiQuery):
        try:
            return q_type.get_query_from_data(query_data)
        except ValueError:
            pass
    return None


def get_query(query_data: str) -> AbiQuery | JsonQuery | None:
    """Fetch the registered query object from telliot-feeds.

    query_data: used to identifying the query
    return: AbiQuery or JsonQuery object
    """
    query_data_bytes = bytes.fromhex(query_data)
    query = get_query_from_data(query_data_bytes)
    return query


def get_feed_from_catalog(tag: str) -> DataFeed | None:
    """Get feed from telliot-feeds mapping if exists."""
    return CATALOG_FEEDS.get(tag)


def get_source_from_data(query_data: bytes, logger: logging) -> DataSource | None:
    """Recreate data source using query type thats decoded from query data field."""
    try:
        query_type, encoded_param_values = decode(["string", "bytes"], query_data)
    except OverflowError:
        logger.error("OverflowError while decoding query data.")
        return None
    try:
        cls = Registry.registry[query_type]
    except KeyError:
        logger.error(f"Unsupported query type: {query_type}")
        return None
    try:
        params_abi = cls.abi
    except AttributeError:
        logger.error(f"query type {query_type} doesn't have abi attirbute to decode params")
        return None
    param_names = [p["name"] for p in params_abi]
    param_types = [p["type"] for p in params_abi]
    param_values = decode(param_types, encoded_param_values)

    feed_builder = DATAFEED_BUILDER_MAPPING.get(query_type)
    if feed_builder is None:
        logger.error(f"query type {query_type} not supported by datafeed builder")
        return None
    source = feed_builder.source
    for key, value in zip(param_names, param_values, strict=False):
        setattr(source, key, value)
    return source


async def get_feed(query_id: str, query: AbiQuery | JsonQuery, logger: logging) -> DataFeed | None:
    """Get the current value for a query from API sources available in telliot-feeds.

    query_id: the hash of the query data used to fetch the value.
    quer: query object that has the source and feed used to get value from relevant API.
    """
    catalog_entry = query_catalog.find(query_id=query_id)
    if len(catalog_entry) == 0:
        source = get_source_from_data(query_data=query.query_data)
        if source is None:
            logger.warning("no source found in telliot feeds found for query")
            return None
        return DataFeed(query=query, source=source)
    else:
        return get_feed_from_catalog(catalog_entry[0].tag)


async def fetch_value(feed: DataFeed) -> OptionalDataPoint:
    """Fetch the value from the data source in telliot-feeds."""
    return await feed.source.fetch_new_datapoint()



import asyncio
import json
import logging
from unittest.mock import MagicMock, patch

from layer_values_monitor.monitor import listen_to_new_report_events

import pytest
import websockets

mock_logger = MagicMock(spec=logging.Logger)


@pytest.mark.asyncio
async def test_websocket_connection(mock_websockets_connect, mock_websocket, event_queue):
    uri = "test-server.com"
    expected_query = json.dumps(
        {
            "jsonrpc": "2.0",
            "method": "subscribe",
            "id": 1,
            "params": {"query": "new_report.reporter_power > 0"},
        }
    )

    mock_websocket.recv.side_effect = websockets.ConnectionClosed(None, None)

    listener_task = asyncio.create_task(listen_to_new_report_events(uri, event_queue, mock_logger))
    await asyncio.sleep(0.1)

    mock_websockets_connect.assert_called_once_with("ws://test-server.com/websocket")
    mock_websocket.send.assert_called_once_with(expected_query)

    listener_task.cancel()
    try:
        await listener_task
    except asyncio.CancelledError:
        pass


@pytest.mark.asyncio
async def test_message_processing(mock_websockets_connect, mock_websocket, event_queue, test_report_messages):
    uri = "test-server.com"
    test_messages = test_report_messages[:3]
    message_index = 0

    async def mock_recv():
        nonlocal message_index
        if message_index < len(test_messages):
            message = test_messages[message_index]
            message_index += 1
            return message
        else:
            raise websockets.ConnectionClosed(None, None)

    mock_websocket.recv.side_effect = mock_recv

    listener_task = asyncio.create_task(listen_to_new_report_events(uri, event_queue, mock_logger))
    await asyncio.sleep(0.1)

    listener_task.cancel()
    try:
        await listener_task
    except asyncio.CancelledError:
        pass

    assert event_queue.qsize() == len(test_messages)

    for expected_message in test_messages:
        message = await event_queue.get()
        assert message == json.loads(expected_message)


@pytest.mark.asyncio
async def test_connection_closed_handling(mock_websockets_connect, mock_websocket, event_queue):
    uri = "test-server.com"

    mock_websocket.recv.side_effect = websockets.ConnectionClosed(None, None)

    with patch("layer_values_monitor.main.logger") as patched_logger:
        listener_task = asyncio.create_task(listen_to_new_report_events(uri, event_queue, patched_logger))
        await asyncio.sleep(0.1)

        listener_task.cancel()
        try:
            await listener_task
        except asyncio.CancelledError:
            pass

        patched_logger.warning.assert_called_with("WebSocket connection closed: no close frame received or sent")

        patched_logger.info.assert_any_call("going through the retry phase since connection was closed")


@pytest.mark.asyncio
async def test_multiple_messages_before_close(mock_websockets_connect, mock_websocket, event_queue, test_report_messages):
    uri = "test-server.com"
    test_messages = test_report_messages

    mock_websocket.recv.side_effect = test_messages + [websockets.ConnectionClosed(None, None)]
    listener_task = asyncio.create_task(listen_to_new_report_events(uri, event_queue, mock_logger))
    await asyncio.sleep(0.1)

    listener_task.cancel()
    try:
        await listener_task
    except asyncio.CancelledError:
        pass
    assert event_queue.qsize() == len(test_messages)

    for expected_message in test_messages:
        message = await event_queue.get()
        assert message == json.loads(expected_message)



"""Values monitor."""

import asyncio
import json
import logging
import random
import time
from typing import Any

from layer_values_monitor.config_watcher import ConfigWatcher
from layer_values_monitor.constants import DENOM
from layer_values_monitor.custom_types import Metrics, Msg, NewReport
from layer_values_monitor.discord import generic_alert
from layer_values_monitor.dispute import (
    determine_dispute_category,
    determine_dispute_fee,
)
from layer_values_monitor.evm_call import get_evm_call_trusted_value
from layer_values_monitor.telliot_feeds import fetch_value, get_feed, get_query
from layer_values_monitor.threshold_config import ThresholdConfig
from layer_values_monitor.utils import add_to_table, get_metric, remove_0x_prefix

import websockets
from telliot_feeds.datafeed import DataFeed


async def listen_to_new_report_events(uri: str, q: asyncio.Queue, logger: logging) -> None:
    """Connect to a layer websocket and fetch new reports and add them to queue for monitoring.

    uri: address to chain (ie ws://localhost:26657/websocket)
    q: queue to store raw response for later processing
    logger:
    """
    logger.info("Starting WebSocket connection...")
    query = json.dumps(
        {
            "jsonrpc": "2.0",
            "method": "subscribe",
            "id": 1,
            "params": {"query": "new_report.reporter_power > 0"},
        }
    )
    uri = f"ws://{uri}/websocket"
    retry_count = 0
    base_delay = 1
    max_delay = 60

    while True:
        try:
            logger.info(f"Connecting to WebSocket at {uri}... (Attempt {retry_count + 1})")
            async with websockets.connect(uri) as websocket:
                await websocket.send(query)
                logger.info("Successfully subscribed to events.")
                # Reset retry count on successful connection
                retry_count = 0
                while True:
                    response = await websocket.recv()
                    await q.put(json.loads(response))

        except websockets.ConnectionClosed as e:
            logger.warning(f"WebSocket connection closed: {e}")
        except websockets.WebSocketException as e:
            logger.error(f"WebSocket error: {e}")
        except Exception as e:
            logger.error(f"Unexpected error: {e}", exc_info=True)
        logger.info("going through the retry phase since connection was closed")
        # Calculate delay with exponential backoff and jitter for reconnection
        delay = min(base_delay * (2**retry_count), max_delay)
        # Add jitter to prevent thundering herd problem
        jitter = random.uniform(0, 0.1 * delay)
        actual_delay = delay + jitter

        retry_count += 1
        logger.info(f"Reconnecting in {actual_delay:.2f} seconds (retry {retry_count})")
        await asyncio.sleep(actual_delay)


async def raw_data_queue_handler(
    raw_data_q: asyncio.Queue,
    new_reports_q: asyncio.Queue,
    logger: logging,
    max_iterations: float = float("inf"),  # use iterations var for testing purposes instead of using a while loop
) -> None:
    """Inspect reports in reports queue and process to see if they should disputed."""
    logger.info("Inspecting reports...")
    iterations = 0
    reports_collections: dict[str, list[NewReport]] = {}
    # current height being collected for
    current_height = 0
    while iterations < max_iterations:
        iterations += 1
        raw_data: dict[str, Any] = await raw_data_q.get()
        raw_data_q.task_done()

        result: dict[str, Any] = raw_data.get("result")
        if result is None:
            continue

        events = result.get("events")
        if events is None:
            continue

        try:
            # get current height from event
            height = events["tx.height"][0]
            height = int(height)
            report = NewReport(
                query_type=events["new_report.query_type"][0],
                query_data=events["new_report.query_data"][0],
                query_id=events["new_report.query_id"][0],
                value=events["new_report.value"][0],
                aggregate_method=events["new_report.aggregate_method"][0],
                cyclelist=events["new_report.cyclelist"][0],
                power=events["new_report.reporter_power"][0],
                reporter=events["new_report.reporter"][0],
                timestamp=events["new_report.timestamp"][0],
                meta_id=events["new_report.meta_id"][0],
                tx_hash=events["tx.hash"][0],
            )
        except (KeyError, IndexError) as e:
            logger.warning(f"malformed report returned by websocket: {e.__str__()}")
            continue

        # check if height has incremented from the reports we are collecting
        # then clear collection and start a new collection
        if height > current_height:
            if len(reports_collections) > 0:
                await new_reports_q.put(dict(reports_collections))
                reports_collections.clear()
            current_height = height

        reports_collected = reports_collections.get(report.query_id, None)
        if reports_collected is None:
            reports_collections[report.query_id] = [report]
        else:
            reports_collections[report.query_id].append(report)


async def inspect_reports(
    reports: list[NewReport],
    disputes_q: asyncio.Queue,
    config_watcher: ConfigWatcher,
    logger: logging,
    threshold_config: ThresholdConfig,
) -> None:
    """Fetch value for a query id and check it against the list of reports' values and determine if any are disputable.

    reports: list of new_reports for a query id
    disputes_q: the queue to add disputes to for dispute processing
    config_watcher: live config for query ids threshold settings
    """
    if len(reports) == 0:
        return None
    query_data = reports[0].query_data
    query_id = reports[0].query_id
    query_type = reports[0].query_type
    _config: dict[str, str] = config_watcher.get_config().get(query_id.lower())
    if _config is None:
        # use globals if no specific config for query id
        metrics = get_metric(
            query_type,
            logger,
            threshold_config,
        )
        # handle not supported query types
        if metrics is None:
            logger.error("no custom configuration and no global thresholds set so can't check value")
            return None
    else:
        metrics = Metrics(
            metric=_config.get("metric"),
            alert_threshold=_config.get("alert_threshold"),
            warning_threshold=_config.get("warning_threshold"),
            minor_threshold=_config.get("minor_threshold"),
            major_threshold=_config.get("major_threshold"),
        )

        if any(
            [
                not metrics.metric,
                not metrics.alert_threshold,
                not metrics.warning_threshold,
                not metrics.minor_threshold,
                not metrics.major_threshold,
            ]
        ):
            logger.error(f"config for {query_id} not set properly")
            return None

    query = get_query(query_data)
    feed = await get_feed(query_id, query, logger)

    if query_type.lower() == "evmcall":
        return await inspect_evm_call_reports(
            reports,
            feed,
            disputes_q,
            query_id,
            metrics,
            logger,
        )

    trusted_value, _ = await fetch_value(feed)
    if trusted_value is None:
        logger.error(f"unable to fetch trusted value for query id: {query_id}")
        return None
    for r in reports:
        reported_value = query.value_type.decode(bytes.fromhex(r.value))
        await inspect(r, reported_value, trusted_value, disputes_q, metrics, logger)

    return None


async def new_reports_queue_handler(
    new_reports_q: asyncio.Queue,
    disputes_q: asyncio.Queue,
    config_watcher: ConfigWatcher,
    logger: logging,
    threshold_config: ThresholdConfig,
) -> None:
    """Handle new reports from the queue and process them."""
    logger.debug("Handling new reports...")
    running_tasks = set()

    while True:
        new_reports: dict = await new_reports_q.get()

        for report in new_reports.values():
            # Create a task for each query id
            task = asyncio.create_task(inspect_reports(report, disputes_q, config_watcher, logger, threshold_config))

            # Add task tracking and cleanup
            running_tasks.add(task)
            task.add_done_callback(running_tasks.discard)

            # Handle exceptions
            def handle_task_exception(t: asyncio.Task) -> None:
                if t.exception():
                    logger.error(f"Task raised exception: {t.exception()}")

            task.add_done_callback(handle_task_exception)

        new_reports_q.task_done()


def is_disputable(
    metric: str, alert_threshold: float, dispute_threshold: float, reported_value: Any, trusted_value: Any, logger: logging
) -> tuple[bool, bool, float] | tuple[None, None, None]:
    """Determine if a value is disputable based on comparison with a trusted value using specified metrics and thresholds."""
    if metric.lower() == "percentage":
        percent_diff: float = (reported_value - trusted_value) / trusted_value
        percent_diff = abs(percent_diff)
        logger.debug(f"percent diff: {percent_diff}, reported value: {reported_value} - trusted value: {trusted_value}")
        if dispute_threshold == 0:
            return percent_diff >= alert_threshold, False, percent_diff
        return percent_diff >= alert_threshold, percent_diff >= dispute_threshold, percent_diff

    if metric.lower() == "equality":
        logger.info(f"checking equality of values, reported value: {reported_value}, trusted value: {trusted_value}")
        # TODO: should this be a string comparison?
        is_not_equal = remove_0x_prefix(str(reported_value)).lower() != remove_0x_prefix(str(trusted_value).lower())
        if dispute_threshold == 0:
            return is_not_equal, False, float(is_not_equal)
        return is_not_equal, is_not_equal, float(is_not_equal)

    if metric.lower() == "range":
        diff = float(abs(reported_value - trusted_value))
        if dispute_threshold == 0:
            return diff >= alert_threshold, False, diff
        return diff >= alert_threshold, diff >= dispute_threshold, diff

    logger.error(f"unsupported metric: {metric}")
    return None, None, None


async def inspect_evm_call_reports(
    reports: list[NewReport],
    feed: DataFeed,
    disputes_q: asyncio.Queue,
    query_id: str,
    metrics: Metrics,
    logger: logging,
) -> None:
    """Inspect reports for evm call query type and check if they are disputable.

    Reason they are different is because we need to specifically fetch values for each report
    because each value needs to be decoded and is dependant on the block timestamp in the value
    to fetch the trusted value from the relevant chain.

    reports: list of new_reports for a query id
    disputes_q: the queue to add disputes to for dispute processing
    query_id: the hash of the query data used to fetch the value.
    metrics: the metrics object for the query id
    logger: the logger object
    feed: the data feed object for the query id
    """
    for r in reports:
        trusted_value = await get_evm_call_trusted_value(r.value, feed)
        if trusted_value is None:
            logger.error(f"unable to fetch trusted value for query id: {query_id}")
            continue
        await inspect(r, r.value, trusted_value, disputes_q, metrics, logger)
    return None


async def inspect(
    report: NewReport,
    reported_value: Any,
    trusted_value: Any,
    disputes_q: asyncio.Queue,
    metrics: Metrics,
    logger: logging,
) -> None:
    """Inspect a new report and check if it is disputable."""
    display = {
        "REPORTER": report.reporter,
        "QUERY_TYPE": report.query_type,
        "QUERY_ID": report.query_id,
        "AGGREGATE_METHOD": report.aggregate_method,
        "CYCLELIST": report.cyclelist,
        "POWER": report.power,
        "TIMESTAMP": report.timestamp,
        "TRUSTED_VALUE": trusted_value,
        "TX_HASH": report.tx_hash,
    }
    display["CURRENT_TIME"] = int(time.time() * 1000)
    display["TIME_DIFF"] = display["CURRENT_TIME"] - (int(report.timestamp))
    display["VALUE"] = reported_value
    # compare values and check against threshold- three metrics(percentage, equality, range)
    alertable, disputable, diff = is_disputable(
        metrics.metric,
        metrics.alert_threshold,
        metrics.warning_threshold,
        reported_value,
        trusted_value,
        logger=logger,
    )
    if alertable is None:
        return None
    if alertable:
        msg = f"found an alertable value. trusted value: {trusted_value}, reported value: {reported_value}"
        logger.info(msg)
        generic_alert(msg + f" tx hash: {report.tx_hash}")

    display["DISPUTABLE"] = disputable

    if disputable and metrics.warning_threshold > 0:
        logger.info(f"found a disputable value. trusted value: {trusted_value}, reported value: {reported_value}")
        # if dispute add message to dispute queue
        category = determine_dispute_category(
            category_thresholds={
                "major": metrics.major_threshold,
                "minor": metrics.minor_threshold,
                "warning": metrics.warning_threshold,
            },
            diff=diff,
        )
        if category is not None:
            fee = determine_dispute_fee(category, int(report.power))
            await disputes_q.put(Msg(report.reporter, report.query_id, report.meta_id, category, fee.__str__() + DENOM))
    add_to_table(display)



import logging
from unittest.mock import MagicMock, patch

from layer_values_monitor.threshold_config import ThresholdConfig
from layer_values_monitor.utils import Metrics, get_metric

import pytest

mock_logger = MagicMock(spec=logging.Logger)


@pytest.fixture
def threshold_values():
    return ThresholdConfig(
        percentage_alert=0.9,
        percentage_warning=0.8,
        percentage_minor=0.7,
        percentage_major=0.6,
        range_alert=0.5,
        range_warning=0.4,
        range_minor=0.3,
        range_major=0.2,
        equality_alert=1.0,
        equality_warning=1.0,
        equality_minor=1.0,
        equality_major=0.0,
    )


def test_spotprice_returns_percentage_metric(threshold_values):
    result = get_metric(query_type="SPOTPRICE", logger=mock_logger, threshold_config=threshold_values)

    expected = Metrics(
        metric="percentage",
        alert_threshold=threshold_values.percentage_alert,
        warning_threshold=threshold_values.percentage_warning,
        minor_threshold=threshold_values.percentage_minor,
        major_threshold=threshold_values.percentage_major,
    )

    assert result == expected


def test_evmcall_returns_equality_metric(threshold_values):
    """Test that EVMCALL query type returns equality metrics"""
    result = get_metric(query_type="EVMCALL", logger=mock_logger, threshold_config=threshold_values)

    expected = Metrics(
        metric="equality",
        alert_threshold=1.0,  # Hardcoded in the function
        warning_threshold=threshold_values.equality_warning,
        minor_threshold=threshold_values.equality_minor,
        major_threshold=threshold_values.equality_major,
    )

    assert result == expected


def test_invalid_query_type(threshold_values):
    with patch("layer_values_monitor.logger.logger.error") as patched_logger:
        mock_logger.error = patched_logger
        result = get_metric(query_type="INVALID_TYPE", logger=mock_logger, threshold_config=threshold_values)

        assert result is None
        patched_logger.assert_called_once_with("query type INVALID_TYPE not found in global metrics")


def test_case_insensitive_query_type(threshold_values):
    # Test with lowercase
    result_lower = get_metric(query_type="spotprice", logger=mock_logger, threshold_config=threshold_values)

    # Test with uppercase
    result_upper = get_metric(query_type="SPOTPRICE", logger=mock_logger, threshold_config=threshold_values)

    # Test with mixed case
    result_mixed = get_metric(query_type="SpOtPrIcE", logger=mock_logger, threshold_config=threshold_values)

    assert result_lower == result_upper
    assert result_lower == result_mixed


def test_range_metric_not_in_enum(threshold_values):
    with patch("layer_values_monitor.logger.logger.error") as patched_logger:
        mock_logger.error = patched_logger
        result = get_metric(query_type="RANGE", logger=mock_logger, threshold_config=threshold_values)

        assert result is None
        patched_logger.assert_called_once_with("query type RANGE not found in global metrics")



"""LVM helper functions."""

import logging
import os

from layer_values_monitor.constants import CSV_FILE, TABLE
from layer_values_monitor.custom_types import GlobalMetric, Metrics
from layer_values_monitor.threshold_config import ThresholdConfig

from pandas import DataFrame


def get_metric(
    query_type: str,
    logger: logging,
    threshold_config: ThresholdConfig,
) -> Metrics | None:
    """Return Metrics/Thresholds given a query type."""
    try:
        metric = GlobalMetric[query_type.upper()].value
    except KeyError:
        logger.error(f"query type {query_type} not found in global metrics")
        print(f"ERROR: query type {query_type} not found in global metrics")
        return None
    if metric == "percentage":
        return Metrics(
            metric=metric,
            alert_threshold=threshold_config.percentage_alert,
            warning_threshold=threshold_config.percentage_warning,
            minor_threshold=threshold_config.percentage_minor,
            major_threshold=threshold_config.percentage_major,
        )
    elif metric == "range":
        return Metrics(
            metric=metric,
            alert_threshold=threshold_config.range_alert,
            warning_threshold=threshold_config.range_warning,
            minor_threshold=threshold_config.range_minor,
            major_threshold=threshold_config.range_major,
        )
    elif metric == "equality":
        return Metrics(
            metric=metric,
            alert_threshold=threshold_config.equality_alert,
            warning_threshold=threshold_config.equality_warning,
            minor_threshold=threshold_config.equality_minor,
            major_threshold=threshold_config.equality_major,
        )


def remove_0x_prefix(s: str) -> str:
    """Remove 0x prefix if there from hex string."""
    if s.startswith("0x"):
        return s[2:]
    return s


def add_to_table(entry: dict[str, str]) -> None:
    """Add entry to table and print it."""
    TABLE.append(entry)
    os.system("clear")
    # df = pd.DataFrame(table)
    df = DataFrame(TABLE).sort_values(by="TIMESTAMP")
    print(df.to_string(index=False, justify="center"))

    if os.path.exists(CSV_FILE):
        # if file exists then just append to it
        DataFrame([entry]).to_csv(CSV_FILE, mode="a", header=False, index=False)
    else:
        # else create file and add header
        DataFrame([entry]).to_csv(CSV_FILE, mode="w", header=True, index=False)



"""Threshold configuration for the Layer Values Monitor."""


class ThresholdConfig:
    """Configuration for global threshold values."""

    def __init__(
        self,
        # percentage thresholds
        percentage_alert: float | None = None,
        percentage_warning: float | None = None,
        percentage_minor: float | None = None,
        percentage_major: float | None = None,
        # range thresholds
        range_alert: float | None = None,
        range_warning: float | None = None,
        range_minor: float | None = None,
        range_major: float | None = None,
        # equality thresholds
        equality_alert: float | None = None,
        equality_warning: float | None = None,
        equality_minor: float | None = None,
        equality_major: float | None = None,
    ) -> None:
        """Initialize the ThresholdConfig with default values."""
        # Set all properties
        self.percentage_alert = percentage_alert
        self.percentage_warning = percentage_warning
        self.percentage_minor = percentage_minor
        self.percentage_major = percentage_major

        self.range_alert = range_alert
        self.range_warning = range_warning
        self.range_minor = range_minor
        self.range_major = range_major

        self.equality_alert = equality_alert
        self.equality_warning = equality_warning
        self.equality_minor = equality_minor
        self.equality_major = equality_major

    @classmethod
    def from_args(cls, args: object) -> "ThresholdConfig":
        """Create a config from parsed command line arguments."""
        if args.use_custom_config:
            return cls()

        # Check if required values are present
        if any(
            [
                args.global_percentage_alert_threshold is None,
                args.global_range_alert_threshold is None,
                args.global_equality_alert_threshold is None,
                # percentage
                args.global_percentage_warning_threshold is None,
                args.global_percentage_minor_threshold is None,
                args.global_percentage_major_threshold is None,
                # range
                args.global_range_warning_threshold is None,
                args.global_range_minor_threshold is None,
                args.global_range_major_threshold is None,
                # equality
                args.global_equality_warning_threshold is None,
                args.global_equality_minor_threshold is None,
                args.global_equality_major_threshold is None,
            ]
        ):
            raise ValueError("Global alert threshold flags required if not using custom config")

        return cls(
            # Percentage thresholds
            percentage_alert=args.global_percentage_alert_threshold,
            percentage_warning=args.global_percentage_warning_threshold,
            percentage_minor=args.global_percentage_minor_threshold,
            percentage_major=args.global_percentage_major_threshold,
            # Range thresholds
            range_alert=args.global_range_alert_threshold,
            range_warning=args.global_range_warning_threshold,
            range_minor=args.global_range_minor_threshold,
            range_major=args.global_range_major_threshold,
            # Equality thresholds
            equality_alert=args.global_equality_alert_threshold,
            equality_warning=args.global_equality_warning_threshold,
            equality_minor=args.global_equality_minor_threshold,
            equality_major=args.global_equality_major_threshold,
        )



"""Type."""

from dataclasses import dataclass
from enum import Enum
from typing import Literal

DisputeCategory = Literal["warning", "minor", "major"]


class GlobalMetric(Enum):
    """Enumeration of global metrics with their associated measurement types."""

    SPOTPRICE = "percentage"
    EVMCALL = "equality"


@dataclass
class NewReport:
    """Store information about a newly submitted report."""

    query_type: str
    query_data: str
    query_id: str
    value: str
    aggregate_method: str
    cyclelist: str
    power: str
    reporter: str
    timestamp: str
    meta_id: str
    tx_hash: str


@dataclass
class Msg:
    """Represent a dispute message."""

    reporter: str
    query_id: str
    meta_id: str
    category: DisputeCategory
    fee: str


@dataclass
class Metrics:
    """Define threshold metrics for setting user configured metrics."""

    metric: str
    alert_threshold: float
    warning_threshold: float
    minor_threshold: float
    major_threshold: float
